package com.mhsoftware.cdaily.support.security;
/*
	connectDaily Web Calendar Software

	Copyright 2016, MH Software, Inc. All Rights Reserved
	5023 W 120th Ave #311, Broomfield CO 80020
	+1 303 438 9585

	Last Edit By: "$Author: gsexton $"
	Last Checkin: "$Date: 2018/03/02 18:42:08 $"
	Revision #  : "$Revision: 1.2 $"

*/
import java.util.List;
import java.util.Locale;
import java.util.ResourceBundle;

import javax.servlet.ServletContext;

import org.mindrot.BCrypt;
import com.MHSoftware.string.StringUtil;
import com.mhsoftware.cdaily.business.User;

/**
 *  This class implements AuthenticationProvider using 
 *  BCrypt to hash the password.
 *  
 *  <br><br>
 *  If the stored password is an MD5SUM, this will transparently
 *  convert the password from an MD5SUM hash to Bcrypt.
 *  
 *  <br><br>
 *  Some things to note:
 *  <ul>
 *  <li>The BCrypt.hashpw() function returns the hash with the
 *  salt embedded into it, so storing the hash separately is not
 *  necessary.
 *  <li>I really wanted to go to Argon2, but there were no Java
 *  implementations available when I wrote this.
 *  </ul>
 *  @since 4.0.6
 *  
 *	@author George Sexton
 */
public class AuthProviderBCrypt extends AuthenticationProvider {

	public AuthProviderBCrypt(ServletContext sc){
		super();
	}

	public LoginResult authenticateUser(String UserID, String Password, String TransformSeed, ResourceBundle rb, Locale lc) {
		throw new UnsupportedOperationException();
	}

	public LoginResult authenticateUser(String UserID, String Password, ResourceBundle rb, Locale lc) {
		lr = new LoginResult();
		User u = User.getUserByName(UserID);
		if (u == null) {
			lr.getFatalList().add(rb.getString("InvalidUserOrPW"));
		} else {
			lr.setUser(u);
			List l = verifyLoginPreconditions(u, rb);
			if (l.isEmpty()) {
				if (!StringUtil.isEmpty(Password) && !transformedMatchesThisProvider(u.password)) {
					AuthenticationProvider md5 = new AuthProviderMD5(null);
					if (md5.transformedMatchesThisProvider(u.password) &&
						md5.verifyPassword(Password, u.password)
					   ) {
						u.password = transformPassword(Password);
					} else if (u.password.equals(Password) && !u.password.startsWith("DISABLED")) {
						// Plain text. 
						u.password=transformPassword(Password);
					}
				}
				if (verifyPassword(Password, u.password)) {
					checkJoeAccount(u,Password,lc);
					processSuccessfulLogin(u, rb, lc);
				} else {
					processFailedLogin(u, rb);
				}
			}
		}
		return lr;
	}

	/**
	 *  Check if this is a Joe Account, and if so, force the
	 *  password expiration (even if password expiration is not
	 *  enabled).
	 *  <br><br>
	 *  For example:
	 *  <br><br>
	 *  User: Joe<br>
	 *  Password: Joe1<br>
	 *  <br><br>
	 *  or the password contains a substring of the username.
	 *  <br><br>
	 *  User: Joe Smith
	 *  Password: Smith
	 *  <br><br>
	 *  
	 * @param u 
	 * @param Password 
	 * @param lc 
	 */
	private void checkJoeAccount(User u,String Password, Locale lc){
		String sUser=u.user_name.toLowerCase(lc),
			sPass=Password.toLowerCase(lc);
		if (sPass.indexOf(sUser)>=0 || sUser.indexOf(sPass)>=0) {
			// this is a joe account. force expire the password.
            /* 
             
            2018-01-10 George Sexton 
             
            A user complained they keep getting the expired message. After looking, 
            I realized because they had a Joe account. I'm not sure why when I 
            originally wrote the checkJoeAccount() method that I purposely allow 
            infinite logins, but evidently I did. On reflection, this is defeating 
            the purpose which is to force a password change. 
             
            Perhaps this was just a logic error on my part that if the system had 
            0 Grace logins, give them a chance to login. 
             
            By now, anyone with a joe account has seen a over year of warnings, 
            so they just need to be dealt with.
             
			if (u.grlogrem<1) {
				u.grlogrem=1;
            } 
            */ 
			this.setForcePasswordExpirationCheck(true); 
		}
	}

	/**
	 * Hash the specified password. The resulting string is in 
	 * BCrypt format, which will have a generated salt built into 
	 * it. 
	 * 
	 * @param Password The plaintext password.
	 * 
	 * @return String BCrypt formatted string. Refer to: 
	 *  
	 * https://en.wikipedia.org/wiki/Bcrypt 
	 *  
	 */
	public String transformPassword(final String Password) {
		String result = null;
		if (Password != null) {
			result = BCrypt.hashpw(Password.trim(), BCrypt.gensalt());
		}
		if (isDebugMode()) {
			System.out.println(getClass() + ".transformPassword(\"" + Password + "\")");
			System.out.println("Returning Password: " + result + " " + (result == null ? "" : result.length() + " Chars"));
		}
		return result;
	}

	public boolean verifyPassword(String plainTextPassword, String transformedPassword) {
		boolean bResult = false;
		try {
			if (!StringUtil.isEmpty(plainTextPassword)) {
				bResult = BCrypt.checkpw(plainTextPassword, transformedPassword); 
			}
		} catch (IllegalArgumentException iae) {
			// An older algorithm...
		}
		return bResult;
	}

	public boolean transformedMatchesThisProvider(String transformed) {
		return transformed.startsWith("$2a$");
	}

	public int getProviderFlags(){
		return AuthenticationProvider.CHANGES_PASSWORDS | AuthenticationProvider.WANT_SSL;
	}
}
